/*
 * Copyright (c) 2009 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbconfig.h"
#include "cdc162.h"
#include "msc162.h"

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )

#ifdef EP_MS_OUT
/*--------------------------------------------------*/
#define SPI_DDR		DDRB
#define SPI_PORT 	PORTB
#define SPI_PIN 	PINB
#define SPI_SS		PB0
#define SPI_SCK		PB1
#define SPI_MOSI	PB2
#define SPI_MISO	PB3

#define SPI_CS_DDR	DDRB
#define SPI_CS_PORT	PORTB
#define SPI_CS		PB4

static inline void spi_enable(void) {
	bit_clear(SPI_CS_PORT,SPI_CS);
}

static inline void spi_disable(void) {
	bit_set(SPI_CS_PORT,SPI_CS);
}

/*--------------------------------------------------*/

#define USE_SWITCH
#define HARD_SPI

void init_spi() {
	bit_set(SPI_CS_PORT, SPI_CS);
	bit_set(SPI_CS_DDR, SPI_CS);

	bit_clear(SPI_DDR, SPI_MISO);
	bit_set(SPI_PORT, SPI_MISO); // pulluped in
	SPI_PORT &=  ~((1<<SPI_SCK) | (1<<SPI_MOSI) | (1<<SPI_SS));
	SPI_DDR |=    ((1<<SPI_SCK) | (1<<SPI_MOSI) | (1<<SPI_SS));
#ifdef HARD_SPI
	SPCR = (1<<SPE)|(1<<MSTR)|(0<<CPOL)|(0<<CPHA)| (3<<SPR0);
	SPSR = (1<<SPI2X);
	/* SPI mode 0 */
	/* SPI CLOCK 1/64 : 16MHz/64 = 250 kHz */
#endif
}

void term_spi() {
#ifdef HARD_SPI
	SPCR = 0;
	SPSR = 0;
#endif
	SPI_DDR &=   ~((1<<SPI_SCK) | (1<<SPI_MOSI) | (1<<SPI_MISO));
	SPI_PORT &=  ~((1<<SPI_SCK) | (1<<SPI_MOSI) | (1<<SPI_MISO));
	bit_clear(SPI_CS_DDR, SPI_CS);
	bit_clear(SPI_CS_PORT, SPI_CS);
}

#define SPI_SPEED_FAST	(0<<SPR0)  // 1/2 = 8 MHz
//#define SPI_SPEED_FAST	(1<<SPR0)  // 1/8 = 2 MHz
#define SPI_SPEED_SLOW	(3<<SPR0)  // 1/64 = 250 kHz 

static inline void spi_speed(uint8_t c) {
#ifdef HARD_SPI
	SPCR &= ~(3<<SPR0);
	SPCR |= (c & (3<<SPR0));
#if 1
	if (c == SPI_SPEED_FAST) {
		SPSR &= ~(1<<SPI2X);
	} else {
		SPSR |= (1<<SPI2X);
	}
#endif
#endif
}

#ifdef HARD_SPI
static inline void spi_send(uint8_t c) {
	SPDR = c;
}

static inline uint8_t spi_recv_send(uint8_t c) {
	uint8_t r;
	while(bit_is_clear(SPSR,SPIF))
		;
	r = SPDR;
	SPDR = c;
	return r;
}

static inline uint8_t spi_recv() {
	while(bit_is_clear(SPSR,SPIF))
		;
	return SPDR;
}
#endif


#ifdef HARD_SPI
static inline uint8_t spi_write(uint8_t c) {
	SPDR = c;
	while(bit_is_clear(SPSR,SPIF))
		;
	return SPDR;
}

#else
static uint8_t spi_write(uint8_t c) {
	uint8_t i,rx;
	i = 0x80;
	rx = 0;
	do {
		if (c & i) {
			bit_set(SPI_PORT, SPI_MOSI);
		} else {
			bit_clear(SPI_PORT, SPI_MOSI);
		}
		bit_set(SPI_PORT, SPI_SCK);
		if (bit_is_set(SPI_PIN, SPI_MISO)) {
			rx |= i;
		}
		bit_clear(SPI_PORT, SPI_SCK);
		i >>= 1;
	} while (i);
	return rx;
}
#endif

static inline uint8_t spi_read(void) {
	return spi_write(0xff);
}

/* Definitions for MMC/SDC command */
#define CMD0    (0)        /* GO_IDLE_STATE */
#define CMD1    (1)        /* SEND_OP_COND (MMC) */
#define CMD8    (8)        /* SEND_IF_COND */
#define CMD9    (9)        /* SEND_CSD */
#define CMD10   (10)       /* SEND_CID */
#define CMD12   (12)       /* STOP_TRANSMISSION */
#define CMD16   (16)       /* SET_BLOCKLEN */
#define CMD17   (17)       /* READ_SINGLE_BLOCK */
#define CMD18   (18)       /* READ_MULTIPLE_BLOCK */
#define CMD23   (23)       /* SET_BLOCK_COUNT (MMC) */
#define CMD24   (24)       /* WRITE_BLOCK */
#define CMD25   (25)       /* WRITE_MULTIPLE_BLOCK */
#define CMD55   (55)       /* APP_CMD */
#define CMD58   (58)       /* READ_OCR */

#define ACMD13  (13)       /* SD_STATUS (SDC) */
#define ACMD23  (23)       /* SET_WR_BLK_ERASE_COUNT (SDC) */
#define ACMD41  (41)       /* SEND_OP_COND (SDC) */

#define CMD_MASK_LBA	0x80
#define CMD_MASK_CONST	0x40

#define CMD55_CONST	( (6 * 0) | CMD_MASK_CONST)
#define CMD0_CONST	( (6 * 1) | CMD_MASK_CONST)
#define CMD8_CONST	( (6 * 2) | CMD_MASK_CONST)
#define ACMD41_CONST	( (6 * 3) | CMD_MASK_CONST)
#define CMD16_CONST	( (6 * 4) | CMD_MASK_CONST)

#define CMD17_LBA	(CMD17 | CMD_MASK_LBA)
#define CMD18_LBA	(CMD18 | CMD_MASK_LBA)
#define CMD24_LBA	(CMD24 | CMD_MASK_LBA)
#define CMD25_LBA	(CMD25 | CMD_MASK_LBA)
#define ACMD23_LBA	(ACMD23 | CMD_MASK_LBA)

static PROGMEM uint8_t mmc_cmd_data[] = {
	CMD55  |0x40,    0,    0,    0,    0, 0x01,
	CMD0   |0x40,    0,    0,    0,    0, 0x95,
	CMD8   |0x40,    0,    0, 0x01, 0xAA, 0x87,
	ACMD41 |0x40, 0x40, 0x00, 0x00, 0x00, 0x00,
	CMD16  |0x40, 0x00, 0x00, 0x01, 0x00, 0x01,
};

#define MMC_TYPE_NONE		0
#define MMC_TYPE_MMC		1
#define MMC_TYPE_SDv1		2
#define MMC_TYPE_SDv2		3
#define MMC_TYPE_MASK		0x7f
#define MMC_TYPE_BLOCK		0x80

uint8_t mmc_type;

uint32u_t mmc_offset;

uint8_t mmc_buf[16];
/* used by mmc_init , mm_read_16 , mmc_read_capacity */

#define MMC_RESP_OK		0
#define MMC_RESP_IDLE		1

/*
 * RETURN VALUE
 *    1xxxxxxx FRAME ERROR
 *    00000000 OK
 *    00000001 Inidle State
 */

#ifdef HARD_SPI
#define MMC_WAIT_MAX	6000		// about 200ms @ 8 MHz bitrate
#define MMC_READ_WAIT_MAX	6000		// ??
#define MMC_WRITE_WAIT_MAX	6000		// ??
#else
#define MMC_WAIT_MAX	1000
#define MMC_READ_WAIT_MAX	3000
#define MMC_WRITE_WAIT_MAX	3000
#endif

static uint16_t mmc_wait_cnt;

//#define MMC_LOG
#ifndef MMC_LOG
#define mmc_debug(ch,f)
#define mmc_debug_enable()
#else
#define MMC_LOG_SIZE	0x80

static uint8_t mmc_log[MMC_LOG_SIZE];
static uint8_t mmc_log_idx;
static uint8_t mmc_log_flag;

static void __mmc_debug(uint8_t ch,int8_t f) {
	if (!f && !mmc_log_flag) return;
	if (mmc_log_idx < MMC_LOG_SIZE) {
		mmc_log[mmc_log_idx++] = ch;

	}
}
#define mmc_debug(ch,f)	__mmc_debug(ch,f)
#define mmc_debug_enable() {mmc_log_flag = 1;}
#endif

#define MSC_MAXLUN	2

static void mmc_command_done() {
	spi_disable();
	spi_read();	// dummy clock;
}

int8_t mmc_command(uint8_t cmd) {
	int8_t i,r;

	mmc_command_done();
	spi_enable();

	if (cmd & CMD_MASK_LBA) {
		cmd &= ~(CMD_MASK_LBA);
		spi_write(cmd | 0x40);

		for (i=3; i>=0; --i) {
			spi_write(mmc_offset.ub[i]);
		}

		spi_write(0x01);
	} else if (cmd & CMD_MASK_CONST) {
		cmd &= ~(CMD_MASK_CONST);
		for (i=0; i<6; i++,cmd++) {
			spi_write(__LPM(mmc_cmd_data+cmd));
		}
	} else {
		spi_write(cmd | 0x40);
		for (i=1; i<6; i++) {
			spi_write(0);
		}
	}

	if (cmd == CMD12) spi_read();

	/* recieve R1 */
	for (i=0; i<10; i++) {
		r = (int8_t)spi_read();
		if (r >=0) break;
	}
	return r;
}

int8_t mmc_command_app(uint8_t cmd) {
	int8_t r;
	r = mmc_command(CMD55_CONST);
	if ((r == MMC_RESP_IDLE) || (r == MMC_RESP_OK)) {
		r = mmc_command(cmd);
	}
	return r;
}

static int8_t mmc_command_init() {
	int8_t j,r;
	uint16_t i;

	for (i=0; i< MMC_WAIT_MAX; i++) {
	    for (j=0; j<32; j++) {
		r = mmc_command_app(ACMD41_CONST);
		if (r == MMC_RESP_OK) goto out;
		if (r != MMC_RESP_IDLE) goto out;
	    }
	}
out:
	return r;
}

uint8_t mmc_init() {
	int8_t i,r;

	mmc_type = MMC_TYPE_NONE;
	init_spi();

	/* dummy clock >= 80 */
	for (i=0; i<10; i++) {
		spi_read();
	}

	r = mmc_command(CMD0_CONST);
//mmc_buf[8] = r;
	if (r != MMC_RESP_IDLE) { // not present ?
		goto err_out;
	}

	r = mmc_command(CMD8_CONST);
//mmc_buf[9] = r;
	if (r != MMC_RESP_IDLE) { // SDSC or MMC
		r = mmc_command_init();
		if ((r != MMC_RESP_OK) && (r != MMC_RESP_IDLE)) {
			goto err_out;
		}
		mmc_type = MMC_TYPE_SDv1;
	} else { // SDv2
		// trailer of R7
		for (i=0; i<4; i++) {
			mmc_buf[i] = spi_read();
		}
		if ((mmc_buf[2] != 0x01) || (mmc_buf[3] != 0xAA)) {
			goto err_out;
		}
		r = mmc_command_init();
		if ((r != MMC_RESP_OK) && (r != MMC_RESP_IDLE)) {
			goto err_out;
		}
		// vdd range 2.7V - 3.6V 
		r = mmc_command(CMD58);
//mmc_buf[11] = r;
		if (r != MMC_RESP_OK) {
			goto err_out;
		}
		for (i=4; i<8; i++) {
			mmc_buf[i] = spi_read();
		}
		mmc_type = MMC_TYPE_SDv2;
		if (mmc_buf[4] & 0x40) mmc_type |= MMC_TYPE_BLOCK;
	}
#if 0
	r = mmc_command(CMD16_CONST);
	if (r != MMC_RESP_OK) {
		goto err_out;
	}
#endif
	spi_speed(SPI_SPEED_FAST);
	return mmc_type;
err_out:
	term_spi();
	return mmc_type;
}

void mmc_term() {
	mmc_type = MMC_TYPE_NONE;
	term_spi();
}

uint8_t mmc_read_16(uint8_t cmd) {
	uint8_t r,i;

	r = 0;
	if (mmc_command(cmd) == MMC_RESP_OK) {
		for (i=0; i<100; i++) {
			r = spi_read();
			if (r == 0xFE) break;
		}
	}
	if (r != 0xFE) {
		mmc_command_done();
		goto out;
	}
	for (i=0; i<16; i++) {
		mmc_buf[i] = spi_read();
	}
	spi_read(); // CRC
	spi_read(); // CRC
out:
	return (r != 0xFE);
}

uint32_t mmc_read_capacity() {
	uint8_t i;
	uint8_t shift_cnt;
	uint32u_t capacity;

	capacity.u32 = 0L;
	if (mmc_read_16(CMD9)) {
		goto out;
	}

	mmc_buf[0] &= 0xc0;
	if (mmc_buf[0] == 0x40) { // CIDv2
		capacity.ub[2] = mmc_buf[7] & 0x3f;
		capacity.ub[1] = mmc_buf[8];
		capacity.ub[0] = mmc_buf[9];
		capacity.u32 += 1;
		capacity.u32 <<= 10;
	} else if (mmc_buf[0] == 0xc0) {
		goto out;
	} else { // CIDv1 or MMC
		shift_cnt = mmc_buf[5] & 0x0f; // [83:80] READ_BL_LEN
		capacity.ub[2] = mmc_buf[6] & 0x03;
		capacity.ub[1] = mmc_buf[7];
		capacity.ub[0] = mmc_buf[8] & 0xc0; // [73:62] C_SIZE 
		// capacity = C_SIZE << 6
		i = (mmc_buf[9] & 0x3) <<1;
		if (mmc_buf[10] & 0x80) {
			i |= 1;
		}
		// i = C_SIZE_MULT
		shift_cnt += (i + 2) - 6;
		capacity.u32 += 0x40;
		capacity.u32 <<= shift_cnt;
		capacity.u32 >>= 9; // 512B
	}
out:
	return capacity.u32;
}

/* --------------------------------------------------------------------*/

#define MMC_WAIT_DATA	1	// WAIT 0xFE;
#define MMC_RECV_DATA	2	// READING 
#define MMC_WAIT_READY	3	// WAIT READY before WRITE
#define MMC_SEND_DATA	4	// WRITING 
#define MMC_SEND_STOP	5	// SEND STOP TOKEN and WAIT BUSY
#define MMC_TRANS_DONE	0x59

static uint8_t mmc_trans_state;

#define SC_VENDOR_ID	'U','S','B','1','6','2',' ',' '
#define SC_PRODUCT_ID	'M','M','C','-','M','S','C',' ',' ',' ',' ',' ',' ',' ',' ',' '
#define SC_REVISION_ID	'0','.','0','2'

static uint8_t mmc_inquiry_data[] PROGMEM = {
	0x00, // DeviceType: Direct-access device , 0x05: CDROM
	0x80, // Medium is removable
	0x03, // Version : SPC (=SBC) (ANSI SCSI level 3 ?)
	0x02, // Response data format
	31,   // Additional Length
	0,0,0, // Reserved
	SC_VENDOR_ID,
	SC_PRODUCT_ID,
	SC_REVISION_ID,
};

static uint8_t mmc_trans_inquiry() {
	while (usbmsc_can_putc(1) && (msc_cbw.cbw.trans_len.ub[0] > 0)) {
		usbmsc_putc(__LPM(mmc_inquiry_data+msc_trans_offset.ub[0]));
		msc_trans_offset.ub[0]++;
		msc_cbw.cbw.trans_len.ub[0] --;
		if (msc_trans_offset.ub[0] >= 36) {
	    		msc_trans_proc = msc_trans_skip;
			break;
		}
	}
	return MSC_TRANS_CONT;
}

static uint8_t mmc_trans_read_capacity() {
	uint32_t capacity;
	capacity = mmc_read_capacity() -1;

	if (usbmsc_can_putc(8)) {
		usbmsc_putc((uint8_t)(capacity>>24));
		usbmsc_putc((uint8_t)(capacity>>16));
		usbmsc_putc((uint8_t)(capacity>>8));
		usbmsc_putc((uint8_t)(capacity));
		usbmsc_putc2(0x00); // SECTOR SIZE
		usbmsc_putc(MSC_SECTOR_SIZE >> 8);
		usbmsc_putc(MSC_SECTOR_SIZE & 0xff);
		msc_cbw.cbw.trans_len.ub[0] -= 8;
		msc_trans_proc = msc_trans_skip;
	}
	return MSC_TRANS_CONT;
}

static uint8_t mmc_cur_lun;

static void mmc_rw_error() {
	msc_cbw.cbw.trans_len.u32 <<= 5;
	msc_trans_proc = msc_trans_skip;
	msc_sense_key[mmc_cur_lun] = SC_SENSE_KEY_MEDIUM_ERROR;
	msc_sense_asc[mmc_cur_lun] = SC_ASC_MEDIUM_NOT_PRESENT;
	msc_unit_state[mmc_cur_lun] = MSC_UNIT_NOT_READY;
}

static uint8_t mmc_trans_read() {
	uint8_t i,r;
#ifdef USE_SWITCH
	switch (mmc_trans_state) {
	case MMC_RECV_DATA:
#else
	if (mmc_trans_state == MMC_RECV_DATA) {
#endif
	    while (usbmsc_can_putc(32) && (msc_cbw.cbw.trans_len.u16[0] > 0)) {
#ifdef HARD_SPI
		spi_send(0);
		for (i=0; i<31; i++) {
			__usbmsc_putc_nc(spi_recv_send(0));
		}
		__usbmsc_putc_nc(spi_recv(0));
#else
		for (i=0; i<32; i++) {
			__usbmsc_putc_nc(spi_read());
		}
#endif
		msc_cbw.cbw.trans_len.u16[0]--;
		if ((msc_cbw.cbw.trans_len.u16[0] & 0xf) == 0) {
			spi_read(); // CRC
			spi_read(); // CRC
			mmc_wait_cnt = 0;
			mmc_trans_state = MMC_WAIT_DATA;
			break;
		}
	    }
	    if (msc_cbw.cbw.trans_len.u16[0] == 0) {
		    mmc_command(CMD12);
		    mmc_command_done();
		    return MSC_TRANS_DONE;
	    }
#ifdef USE_SWITCH
	    break;
	case MMC_WAIT_DATA: // WAIT 0xFE;
#else
	} else if (mmc_trans_state == MMC_WAIT_DATA) {
#endif
	    for (i=0; i<32; i++) {
		    r = spi_read();
		    if (r == 0xFE) break;
	    }
	    if (r == 0xFE) {
		    mmc_trans_state = MMC_RECV_DATA;
	    } else {
		    mmc_wait_cnt++;
		    if (mmc_wait_cnt > MMC_READ_WAIT_MAX) {
		        mmc_command(CMD12);
			mmc_rw_error();
			msc_sense_asc[mmc_cur_lun] = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION ;
				;
		    }
	    }
#ifdef USE_SWITCH
	    break;
	default:
	    break;
#endif
	}
	return MSC_TRANS_CONT;
}

static uint8_t mmc_trans_write() {
	uint8_t i,r;
#ifdef USE_SWITCH
	switch (mmc_trans_state) {
	case MMC_SEND_DATA:
#else
	if (mmc_trans_state == MMC_SEND_DATA) {
#endif
	    while (usbmsc_can_getc(32) && (msc_cbw.cbw.trans_len.u16[0] > 0L)) {
#ifdef HARD_SPI
		spi_send(__usbmsc_getc_nc()); 
		for (i=0; i< 31; i++) {
			spi_recv_send(__usbmsc_getc_nc()); 
		}
		spi_recv(); 
#else
		for (i=0; i< 32; i++) {
			spi_write(__usbmsc_getc_nc()); 
		}
#endif
		msc_cbw.cbw.trans_len.u16[0] --;
		if ((msc_cbw.cbw.trans_len.u16[0] & 0xf) == 0) {
			spi_write(0xff); // dummy CRC
			spi_write(0xff); // dummy CRC
			r = spi_read();  // Receive data response 
			if ((r & 0x1f) != 0x05) { // not accepted
				mmc_debug(0xff, 1);
				mmc_debug(r, 1);
				mmc_rw_error();
				msc_sense_asc[mmc_cur_lun] = SC_ASC_LOGICAL_UNIT_NOT_READY;
			} else {
				mmc_wait_cnt = 0;
		    		mmc_trans_state = MMC_WAIT_READY;
				mmc_debug(mmc_trans_state, 1);
			}
		        break;
		}
	    }
#ifdef USE_SWITCH
	    break;
	case MMC_WAIT_READY: // WAIT 0xFF;
	case MMC_SEND_STOP:
#else
	} else {
#endif
	    spi_read();
	    for (i=0; i<31; i++) {
		r = spi_read();
		if (r == 0xFF) break;
	    }
	    if (r == 0xFF) {
		if (mmc_trans_state == MMC_SEND_STOP) {
			mmc_debug_enable();
			mmc_debug(0x59,1);
			mmc_command_done();
	    		return MSC_TRANS_DONE;
		}
		if (msc_cbw.cbw.trans_len.u16[0] == 0) {
	    		spi_write(0xFD); // send TOKEN
			mmc_debug(0xFD, 1);
			mmc_wait_cnt = 0;
		    	mmc_trans_state = MMC_SEND_STOP;
		} else {
			spi_write(0xFC); // send TOKEN
			mmc_debug(0xFC, 1);
		    	mmc_trans_state = MMC_SEND_DATA;
		}
	    } else {
		    mmc_wait_cnt++;
		    if (mmc_wait_cnt > MMC_WRITE_WAIT_MAX) {
			mmc_rw_error();
		    }
	    }
#ifdef USE_SWITCH
	    break;
#endif
	}
	return MSC_TRANS_CONT;
}

static void mmc_devcall(uint8_t sc_lun, uint8_t sc_cmd) {
	uint8_t r;

	mmc_debug_enable();
	mmc_debug(sc_cmd, 0);
#ifdef USE_SWITCH
	switch (sc_cmd) {
	case SC_READ_10:
	case SC_WRITE_10:
#else
	if (sc_cmd == SC_READ_10 || sc_cmd == SC_WRITE_10) {
#endif
		if (!(mmc_type & MMC_TYPE_BLOCK)) {
			msc_trans_offset.u32 <<= MSC_SECTOR_SHIFT;
		}
		mmc_wait_cnt = 0;
		mmc_cur_lun = sc_lun;
		if (sc_cmd == SC_READ_10) {
			mmc_offset.u32 = msc_trans_offset.u32;
			mmc_command(CMD18_LBA);
			mmc_trans_state = MMC_WAIT_DATA;
			msc_trans_proc = mmc_trans_read;
		} else {
			mmc_debug(sc_cmd, 1);
			mmc_offset.u16[1] = 0;
			mmc_offset.u16[0] = msc_cbw.cbw.trans_len.ub[0] >> (MSC_SECTOR_SHIFT - 5);
			r = mmc_command_app(ACMD23_LBA);
			mmc_debug(r, 1);
			mmc_offset.u32 = msc_trans_offset.u32;
			r = mmc_command(CMD25_LBA);
			mmc_debug(r, 1);
			//spi_write(0xFC); // send TOKEN
			mmc_trans_state = MMC_WAIT_READY;
			mmc_debug(mmc_trans_state, 1);
			msc_trans_proc = mmc_trans_write;
		}
#ifdef USE_SWITCH
		break;
	case SC_INQUIRY:
#else
	} else if (sc_cmd == SC_INQUIRY) {
#endif
		msc_trans_proc = mmc_trans_inquiry;
#ifdef USE_SWITCH
		break;
	case SC_READ_CAPACITY:
#else
	} else if (sc_cmd == SC_READ_CAPACITY) {
#endif
		msc_trans_proc = mmc_trans_read_capacity;
#ifdef USE_SWITCH
		break;
	case SC_TEST_UNIT_READY:
#else
	} else if (sc_cmd == SC_TEST_UNIT_READY) {
#endif
		if (msc_unit_state[sc_lun] == MSC_UNIT_NOT_READY) {
			if (mmc_type) {
				msc_unit_state[sc_lun] = MSC_UNIT_CHANGE_TO_READY;
			} else {
				mmc_init();
			}
			if (msc_unit_state[sc_lun] != MSC_UNIT_NOT_READY) {
				mmc_debug(sc_cmd, 1);
				mmc_debug(mmc_type, 1);
				mmc_debug(msc_unit_state[sc_lun] | 0x80, 1);
			}
		} else if (msc_unit_state[sc_lun] == MSC_UNIT_READY) {
#if 1
			if (mmc_read_16(CMD9)) {
#else
			if (mmc_command(CMD16_CONST) != MMC_RESP_OK) {
#endif
				msc_unit_state[sc_lun] = MSC_UNIT_NOT_READY;
				mmc_term();
				mmc_debug(sc_cmd, 1);
				mmc_debug(mmc_type, 1);
				mmc_debug(msc_unit_state[sc_lun] | 0x80, 1);
			}
		}
#ifdef USE_SWITCH
		break;
	default:
		break;
#endif
	}
}

void msc_mmc_init() {
	mmc_init();
	msc_devattach(mmc_devcall);
}

void mmc_print_status() {
#ifdef MMC_LOG
#if defined(EP_MS_OUT)
	eeprom_write_block (mmc_log, (void *)0x100,MMC_LOG_SIZE);
	eeprom_write_block (mmc_buf, (void *)0xf0,16);
#endif
#if defined(EP_TX) && defined(EP_MS_OUT)
	uint8_t i,j,d;
	uint16_t offset = 0;
	usbcdc_put_pstr(PSTR("mmc "));
	usbcdc_put_hex(mmc_type);
	usbcdc_put_pstr(PSTR("\r\n"));
	for(i=0; i<MMC_LOG_SIZE/16; i++) {
		d = (offset >> 8) & 0xff;
		usbcdc_put_hex(d);
		d = offset & 0xff;
		usbcdc_put_hex(d);
		usbcdc_putc(':');
		for (j=0; j<16; j++) {
			usbcdc_putc(' ');
			d = mmc_log[offset];
			usbcdc_put_hex(d);
			offset++;
		}
		usbcdc_put_pstr(PSTR("\r\n"));
	}
#endif
	mmc_log_idx = 0;
	memset(mmc_log,0,MMC_LOG_SIZE);
#endif
}

#endif
/*------------------------------

OCR : 32 bit data
   bit 0 - 23 VDD Voltage Window 
        bit 15  (2.7 - 2.8 )
        bit 23  (3.5 - 3.6 )
   bit 30 Card Capacity Status (CCS) (1: OK)  
   bit 31 Card Power up status bit (busy) ( 1: busy ) 

CID : 128 bit data

   8bit  Manufacture ID
  16bit  OEM/Application ID (ASCII)
  40bit  Product Name (ASCII)
   8bit  Product Revision
  32bit  Product Serial Number
  12bit  Manufacturing date  
   8bit  CRC+1

CSD : 128bit data

   8bit  CSD_STRUCTURE ( 0x00 : CSDv1 , 0x40 : CSDv2 )

CSDv1 : 

   12bit [73:62] C_SIZE 
    3bit [49:47] C_SIZE_MULT
    4bit [83:80] READ_BL_LEN

    1bit [13:13] PARM_WRITE_PROTECT
    1bit [12:12] TMP_WRITE_PROTECT


  BLOCK_SIZE = 2 ^ READ_BL_LEN
  BLOCKNR = (C_SIZE+1) x (2 ^ C_SIZE_MULT) x 4 

CSDv2 : 
  22 bit [69:48] C_SIZE ( [ 71:70] = 00b )

  CAPACITY = (C_SIZE+1) x 512 KB
--------------------------------------*/

